#include <stdint.h>

// Mode sense data
static const uint8_t MODE_SENSE_6_DATA[8] = {
  0x00,
  0x00,
  0x00,
  0x00,
  0x00,
  0x00,
  0x00,
  0x00
};


// Standard Inquiry Data
static const uint8_t STD_INQUIRY_DATA[36] = {
  0x00, // Direct access block device
  0x80, // RMB bit set to one indicates that the medium is removable
  0x00, // ISO(7..6) ECMA(5..3) ANSI(2..0) Version
  0x02, // Response Data Format
  0x1F, // Additional length (31)
  0x00, // Reserved
  0x00, // Reserved
  0x00, // Reserved
  'M', 'I', 'K', 'R', 'O', 'E', ' ', ' ', // Vendor Information
  'm', 'i', 'c', 'r', 'o', 'S', 'D', ' ', // Product identification
  'F', 'l', 'a', 's', 'h', ' ', ' ', ' ',
  '1', '.', '0', '0'                      // Product Revision Level n.nn
};

// USB Mass storage Page 0 Inquiry Data
static const uint8_t UNIT_SERIAL_NUMBER[] = {
  0x00, // Peripheral qualifier[7..5] device type[4..0]
  0x80, // Page Code 80h
  0x00, // Reserved
  0x03, // Page Length
  0x00, // Product Serial Nnumber
  0x00,
  0x00
};

// Storage callbacks
static uint8_t StorageInit();
static uint8_t StorageIsReady();
static uint8_t StorageIsWriteProtected();
static uint8_t StorageGetCapacity(uint32_t* blockNum, uint32_t *blockSize);
static uint8_t StorageRead(uint8_t *buffer, uint32_t lba, uint16_t blockNum);
static uint8_t StorageWrite(uint8_t *buffer, uint32_t lba, uint16_t blockNum);
static uint8_t* StorageGetInquiryData(uint8_t vpd);
static uint8_t* StorageGetStdInquiryData();
static uint8_t* StorageGetModeSenseData();

typedef struct {
  uint8_t(*StorageInit)();
  uint8_t(*StorageIsReady)();
  uint8_t(*StorageIsWriteProtected)();
  uint8_t(*StorageGetCapacity)(uint32_t *blockNum, uint32_t * blockSize);
  uint8_t(*StorageRead) (uint8_t *buffer, uint32_t lba, uint16_t blockNum);
  uint8_t(*StorageWrite)(uint8_t *buffer, uint32_t lba, uint16_t blockNum);
  uint8_t * (*StorageGetInquiryData)(uint8_t vpd);
  uint8_t * (*StorageGetStdInquiryData)();
  uint8_t * (*StorageGetModeSenseData)();
} TMSCStorageCB;

TMSCStorageCB USBDev_MSCStorageCB = {
  StorageInit,
  StorageIsReady,
  StorageIsWriteProtected,
  StorageGetCapacity,
  StorageRead,
  StorageWrite,
  StorageGetInquiryData,
  StorageGetStdInquiryData,
  StorageGetModeSenseData
};

// STORAGE IMPLEMENTATION

////////////////////////////////////////////////////////////////////////////////
// This is a wrapper for retrieving number of sector device implements.
// It is used to get sector count of the device (for fat formatting purpose,
// assumed sector size is 512 bytes).

// MMC GetMmcSectorCount return codes
static const uint8_t  MMC_OK     =   0,
                      MMC_ERROR  =   255;
                      
static uint8_t GetMmcSectorCount(uint32_t *scCnt)
{
    uint8_t  csdbuf[16];
    uint16_t c_size, c_size_mult, mult,
           read_bl_len, block_len;
    uint32_t size, blocknr;

    // determine MMC/SD card size in MB
    if (Mmc_Read_Csd(csdbuf) != MMC_OK)
    {
        return MMC_ERROR;
    }

    // is it version 2.0?
    if (1 == ((csdbuf[0] & 0xC0) >> 6))
    {
        size  = 0;                  size <<= 8;
        size += csdbuf[7] & 0x3F;   size <<= 8;
        size += csdbuf[8];          size <<= 8;
        size += csdbuf[9];          size <<= 0;

        // size is in 0.5MB, get size in sectors (assumed 512 bytes sector size)
        size *= 1024;
    }
    // if not, it's version 1.xx
    else
    {
        c_size      = ((csdbuf[8] & 0xC0) >> 6) +
                      ((unsigned) csdbuf[7] << 2) +
                      (((unsigned) csdbuf[6] & 0x03) << 10);
        c_size_mult = (csdbuf[10] & 0x80) +
                      (((unsigned) csdbuf[9] & 0x03) << 8);
        c_size_mult = c_size_mult >> 7;

        read_bl_len = csdbuf[5] & 0x0f;

        mult = 1;
        mult = mult << (c_size_mult + 2);

        blocknr = (c_size + 1) * (long) mult;
        block_len = 1;
        block_len = block_len << read_bl_len;

        size = block_len * blocknr;

        // size is in 1B, get size in sectors (assumed 512 bytes sector size)
        size /= 512;
    }

    *scCnt = size;

    return MMC_OK;
}

static uint8_t  storageInitStatus;   // storage initializtion status
// Initializing storage
static uint8_t StorageInit() {
  // initialize a MMC card
  storageInitStatus = Mmc_Init();
  return storageInitStatus;
}

// Get storage capacity, number of blocks and block size
static uint8_t StorageGetCapacity(uint32_t* blockNum, uint32_t *blockSize) {
  GetMmcSectorCount(blockNum);
  *blockSize = 512;
  return 0;
}

// Read storage to buffer
static uint8_t StorageRead(uint8_t *buffer, uint32_t lba, uint16_t blockNum) {
  uint8_t status;
  status = 0;
  Mmc_Multi_Read_Start(lba);
  while (blockNum) {
    Mmc_Multi_Read_Sector(buffer);
    buffer += 512;
    blockNum--;
  }
  Mmc_Multi_Read_Stop();
  return status;
}

// Return storage status
static uint8_t StorageIsReady() {
  if(storageInitStatus)
    return 0; // storage is not ready
  else
    return 1;  // storage is ready
}

// Write to storage
static uint8_t StorageWrite(uint8_t *buffer, uint32_t lba, uint16_t blockNum) {
  uint8_t status;
  status = 0;
  while (blockNum) {
    status |= Mmc_Write_Sector(lba, buffer);
    lba++;
    buffer += 512;
    blockNum--;
  }
  return status;
}

// Get storage protection status
static uint8_t StorageIsWriteProtected() {
  return 0;
}

// Return storage inquiry data
static uint8_t* StorageGetInquiryData(uint8_t vpd) {
  if(vpd == 0x80)
  return (uint8_t *)UNIT_SERIAL_NUMBER;
}

// Get standard inquiry data
static uint8_t* StorageGetStdInquiryData() {
  return (uint8_t *)STD_INQUIRY_DATA;
}

// Get mode sense data
static uint8_t* StorageGetModeSenseData() {
  return (uint8_t *)MODE_SENSE_6_DATA;
}

// END OF STORAGE IMPLEMENTATION